;;; howm-date.el --- Wiki-like keyword jump between files
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-date.el,v 1.7 2004/01/23 11:48:59 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Date

(defun howm-insert-date ()
  (interactive)
  (let ((date (format-time-string howm-date-format)))
    (insert (format howm-insert-date-format date))
    (save-excursion
      (backward-char 11)
      (howm-action-lock-date date howm-insert-date-future))))

(defun howm-action-lock-date (date &optional new)
  (let* ((ret (if new "ok" "goto"))
         (prompt (format "[%s] RET(%s), +num(shift), yymmdd(set), ~yymmdd(repeat), .(today): "
                         (howm-datestr-day-of-week date)
                         ret))
         (c (read-from-minibuffer prompt)))
    (cond 
     ((string= c "")
      (when (not new)
        (howm-action-lock-date-search date)))
     ((string-match "^[-+][0-9]+$" c)
      (howm-action-lock-date-shift (string-to-number c) date))
     ((string-match "^[0-9]+$" c)
      (howm-action-lock-date-set c date new))
     ((string-match "^~\\([0-9]+\\)$" c)
      (howm-action-lock-date-repeat (match-string-no-properties 1 c) date))
     ((string-match "^[.]$" c)
      (howm-action-lock-date-set (format-time-string "%Y%m%d") date))
     (t (error (format "Can't understand %s." c))))))

(defun howm-action-lock-date-search (date)
  (howm-view-search date (howm-search-path) nil nil t)
  (howm-list-normalize))

(defun howm-action-lock-date-shift (n date)
  (howm-datestr-replace (howm-datestr-shift date 0 0 n)))

;; (defun howm-action-lock-date-shift (shift-str date)
;;   (let ((n (string-to-number shift-str)))
;;     (howm-datestr-replace (howm-datestr-shift date 0 0 n))))

(defun howm-action-lock-date-set (val date &optional future-p)
  (howm-datestr-replace (howm-datestr-expand val date future-p)))

(defvar howm-action-lock-date-repeat-max 200)
(defun howm-action-lock-date-repeat (until date)
;;   (let* ((until (progn
;;                   (string-match "^~\\([0-9]+\\)" command)
;;                   (string-to-number (match-string-no-properties 1 command))))
;;          (every (read-from-minibuffer "Every? [RET(all), num(days), w(week), m(month), y(year)] ")))
  (let ((every (read-from-minibuffer "Every? [RET(all), num(days), w(week), m(month), y(year)] ")))
    (let ((max-d (howm-datestr-expand until date t))
          (offset-y (if (string= every "y") 1 0))
          (offset-m (if (string= every "m") 1 0))
          (offset-d (or (cdr (assoc every '(("y" . 0) ("m" . 0) ("w" . 7))))
                        (max (string-to-number every) 1))))
      (let ((d date)
            (i 0)
            (check t))
        (catch 'too-many
          (while (progn
                   (setq d (howm-datestr-shift d offset-y offset-m offset-d))
                   (howm-datestr<= d max-d))
            (when (and check (>= i howm-action-lock-date-repeat-max))
              (if (y-or-n-p (format "More than %d lines. Continue? " i))
                  (setq check nil)
                (throw 'too-many nil)))
            (howm-duplicate-line)
            (howm-datestr-replace d)
            (setq i (+ i 1))))))))

(defun howm-make-datestr (y m d)
  (let ((ti (encode-time 0 0 0 d m y)))
    (format-time-string howm-date-format ti)))

(defun howm-datestr-parse (date)
  (string-match howm-date-regexp date)
  (mapcar (lambda (pos)
            (string-to-number (match-string-no-properties pos date)))
          (list howm-date-regexp-year-pos
                howm-date-regexp-month-pos
                howm-date-regexp-day-pos)))

(defun howm-datestr-to-time (date)
  (let* ((ymd (howm-datestr-parse date))
         (y (car ymd))
         (m (second ymd))
         (d (third ymd)))
    (encode-time 0 0 0 d m y)))

(defun howm-datestr-day-of-week (date)
  (format-time-string "%a" (howm-datestr-to-time date)))

(defun howm-datestr-expand (date base &optional future-p)
  (let* ((base-ymd (howm-datestr-parse base))
         (nval (format "%8s" date))
         (given-ymd (mapcar (lambda (r)
                              (string-to-number (substring nval
                                                           (car r) (second r))))
                            '((0 4) (4 6) (6 8))))
         (carry nil) ;; to force future date
         (dmy (mapcar* (lambda (ox nx)
                         (when future-p
                           (when (and carry (= nx 0))
                             (setq ox (+ ox 1)))
                           (setq carry
                                 (cond ((= nx 0) nil)
                                       ((= nx ox) carry)
                                       ((< nx ox) t)
                                       (t nil))))
                         (if (= nx 0) ox nx))
                       (reverse base-ymd) (reverse given-ymd)))
         (d (car dmy))
         (m (second dmy))
         (y (third dmy)))
    (howm-make-datestr (if (<= y 99) (+ y 2000) y) m d)))

(defun howm-datestr-shift (date y m d)
  (let* ((ymd (howm-datestr-parse date))
         (oy (car ymd))
         (om (second ymd))
         (od (third ymd)))
    (howm-make-datestr (+ oy y) (+ om m) (+ od d))))

(defun howm-datestr<= (date1 date2)
  (or (string< date1 date2)
      (string= date1 date2)))

(defun howm-datestr-replace (date)
  (let ((p (point)))
    (while (not (looking-at howm-date-regexp))
      (backward-char))
    (replace-match date t t)
    (goto-char p)))

(defun howm-duplicate-line ()
  (let ((c (current-column))
        (s (buffer-substring (line-beginning-position) (line-end-position))))
    (end-of-line)
    (insert "\n" s)
    (move-to-column c)))

(provide 'howm-date)

;;; howm-date.el ends here
