;;; howm-view.el --- Wiki-like keyword jump between files
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-view.el,v 1.140 2004/02/18 10:34:15 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

;;; variables

;; customize
(defvar howm-view-pop-up-windows t
  "If non-nil, Override pop-up-windows.")
(defvar howm-view-split-horizontally nil
  "If t, split window horizontally to show summary and contents.
If nil, split vertically. Otherwise, split horizontally at only first time.")
(defvar howm-view-summary-persistent t
  "If non-nil, keep summary buffer on howm-view-summary-open.
If it is a function, the evaluated value is used instead of itself.")
(defvar howm-view-contents-persistent nil
  "If non-nil, keep contents buffer on howm-view-contents-open.
If it is a function, the evaluated value is used instead of itself.")
(defvar howm-view-contents-limit nil
  "Max length for howm-view-contents. Nil means no limit.")
; (defvar howm-view-contents-limit (* 80 200))
(defvar howm-view-summary-name "*howmS*"
  "Format string of buffer name for summary.
%s is replaced with searched string. See `format'.")
(defvar howm-view-contents-name "*howmC*"
  "Format string of buffer name for contents.
%s is replaced with searched string. See `format'.")
;; (defvar howm-view-summary-name "*howmS:%s*"
;;   "Format string of buffer name for summary.
;; %s is replaced with searched string. See `format'.")
;; (defvar howm-view-contents-name "*howmC:%s*"
;;   "Format string of buffer name for contents.
;; %s is replaced with searched string. See `format'.")
(defvar howm-view-summary-mode-hook nil)
(defvar howm-view-contents-mode-hook nil)
(defvar howm-view-open-hook nil)
(defvar howm-view-before-open-hook nil)
(defvar howm-view-hilit-face 'howm-view-hilit-face)
(defface howm-view-hilit-face '((t (:foreground "red")))
  "Face for matched word.")
(defvar howm-view-contents-font-lock-keywords nil)
(defvar howm-view-summary-format
  (let* ((path (format-time-string howm-file-name-format))
         (width (length (file-name-nondirectory path))))
    (concat "%-" (format "%s" width) "s | ")))
(defvar howm-view-header-format
  "\n==========================>>> %s\n"
  "Format string of header for howm-view-contents.
%s is replaced with file name. See `format'.")
(defvar howm-view-header-regexp "^==========================>>> .*$")
(defvar howm-view-search-recenter 5)
(defvar howm-view-open-recenter howm-view-search-recenter)
(defvar howm-view-title-header "=")
;; howm-view-title-regexp is assumed to have a form "^xxxxxxx$"
(defvar howm-view-title-regexp (format "^%s\\( +\\(.*\\)\\|\\)$"
                                     (regexp-quote howm-view-title-header)))
(defvar howm-view-title-regexp-pos 2)
(defvar howm-view-title-regexp-grep (format "^%s +"
                                     (regexp-quote howm-view-title-header)))

(defvar howm-view-use-grep nil)
(defvar howm-view-grep-command "egrep")
(defvar howm-view-fgrep-command "fgrep")
(defvar howm-view-grep-option "-Hnr")
(defvar howm-view-grep-ignore-case-option "-i")
(defvar howm-view-grep-expr-option "-e")

(defvar howm-view-sort-methods
  '(("name" . howm-view-sort-by-name)
    ("name-match" . howm-view-sort-by-name-match)
    ("summary" . howm-view-sort-by-summary)
    ("summary-match" . howm-view-sort-by-summary-match)
;     ("atime" . howm-view-sort-by-atime) ;; nonsense
;     ("ctime" . howm-view-sort-by-ctime) ;; needless
    ("mtime" . howm-view-sort-by-mtime)
    ("date" . howm-view-sort-by-reverse-date)
    ("reminder" . howm-view-sort-by-reminder)
    ("reverse" . howm-view-sort-reverse)))

(defvar howm-view-filter-methods
  '(("name" . howm-view-filter-by-name)
    ("summary" . howm-view-filter-by-summary)
    ("mtime" . howm-view-filter-by-mtime)
;     ("ctime" . howm-view-filter-by-ctime) ;; needless
    ("date" . howm-view-filter-by-date)
    ("reminder" . howm-view-filter-by-reminder)
    ("contents" . howm-view-filter-by-contents)
;     ("uniq" . howm-view-filter-uniq))
  ))

;; referred only when howm-view-use-grep is nil
(defvar howm-view-watch-modified-buffer t)

(defvar howm-view-update-search-ring nil)

;; internal
(defvar howm-view-name nil)
(defvar howm-view-item-list nil)
(defvar howm-view-summary-last-line nil)
(defvar howm-view-contents-end nil)
(make-variable-buffer-local 'howm-view-name)
(make-variable-buffer-local 'howm-view-item-list)
; update contents when changed
(make-variable-buffer-local 'howm-view-summary-last-line)
; end points of items
(make-variable-buffer-local 'howm-view-contents-end)

(defvar *howm-view-summary-check* t)

(defvar *howm-view-font-lock-keywrods* nil)

(defun howm-view-restore-window-configuration ()
  (howm-view-refresh-window-configuration))

(defun howm-view-refresh-window-configuration ()
  (if (member major-mode
              '(howm-view-summary-mode howm-view-contents-mode))
      (howm-view-setup-window-configuration)
    (delete-other-windows)))

(defvar howm-view-window-initializer 'delete-other-windows)
;; (defvar howm-view-window-initializer '(lambda () (pop-to-buffer nil)))
(defun howm-view-setup-window-configuration ()
  (let ((orig (current-buffer))
        (s (howm-view-summary-buffer))
        (c (howm-view-contents-buffer)))
    (when (functionp howm-view-window-initializer)
      (funcall howm-view-window-initializer))
    (switch-to-buffer c)
    (if howm-view-split-horizontally
        (split-window-horizontally)
      (split-window))
    (switch-to-buffer s)
    (select-window (get-buffer-window orig))))

;;; keymap

(defvar howm-view-summary-mode-map nil)
(defvar howm-view-contents-mode-map nil)

(defun howm-view-define-common-key (keymap)
  (let ((m keymap))
;;     (define-key m "?" 'howm-view-help)
    (define-key m "f" 'howm-view-filter)
    (define-key m "s" 'howm-view-filter-by-contents)
    (define-key m "S" 'howm-view-sort)
    (define-key m "R" 'howm-view-sort-reverse)
    (define-key m "q" 'howm-view-kill-buffer)
    (define-key m "X" 'howm-view-dired)
    (define-key m "n" 'next-line)
    (define-key m "p" 'previous-line)
    (define-key m "?" 'describe-mode)
    ))

(when (not howm-view-summary-mode-map)
  (let ((m (make-sparse-keymap)))
    (define-key m "\C-m" 'howm-view-summary-open)
    (define-key m "\C-j" 'howm-view-summary-open)
    (define-key m " " 'howm-view-pop-or-scroll-other-window)
    (define-key m [backspace] 'scroll-other-window-down)
    (define-key m "\C-h" 'scroll-other-window-down)
    (define-key m "j" 'howm-view-scroll-other-window)
    (define-key m "k" 'howm-view-scroll-other-window-down)
    (define-key m "@" 'howm-view-summary-to-contents)
    (define-key m "u" 'howm-view-filter-uniq)
    (define-key m "!" 'howm-view-summary-shell-command)
    (define-key m "0" 'howm-view-summary-to-contents)
    (define-key m "1" 'delete-other-windows)
    (define-key m "2" 'howm-view-pop-window)
    (define-key m "v" 'howm-view-toggle-window)
;;     (define-key m howm-reminder-quick-check-key 'howm-reminder-quick-check)
;;     (define-key m ";" 'howm-view-invoke-action-lock)
    (howm-view-define-common-key m)
    (setq howm-view-summary-mode-map m)))

(when (not howm-view-contents-mode-map)
  (let ((m (make-sparse-keymap)))
    (define-key m "\C-m" 'howm-view-contents-open)
    (define-key m "\C-j" 'howm-view-contents-open)
    (define-key m " " 'scroll-up)
    (define-key m [backspace] 'scroll-down)
    (define-key m "\C-h" 'scroll-down)
    (define-key m "j" 'howm-view-scroll-up)
    (define-key m "k" 'howm-view-scroll-down)
    (define-key m "@" 'howm-view-contents-to-summary)
    (define-key m "0" 'howm-view-contents-to-summary)
    (define-key m "\C-t" 'howm-view-contents-goto-next-item)
    (define-key m "\M-\C-t" 'howm-view-contents-goto-previous-item)
    (define-key m [tab] 'howm-view-contents-goto-next-item)
    (define-key m [(meta tab)] 'howm-view-contents-goto-previous-item)
    (howm-view-define-common-key m)
    (setq howm-view-contents-mode-map m)))

;;; item

;; item = (filename summary place offset home)
(defun howm-view-make-item (filename &optional summary place offset home
                                     privilege)
  (list filename summary place offset home privilege))
(defun howm-view-item-filename      (item) (howm-view-item-nth 0 item ""))  
(defun howm-view-item-summary       (item) (howm-view-item-nth 1 item ""))  
(defun howm-view-item-place         (item) (howm-view-item-nth 2 item nil)) 
(defun howm-view-item-offset        (item) (howm-view-item-nth 3 item nil)) 
(defun howm-view-item-home          (item) (howm-view-item-nth 4 item nil)) 
(defun howm-view-item-privilege     (item) (howm-view-item-nth 5 item nil)) 
(defun howm-view-item-nth (n item default)
  (or (nth n item) default))
(defun howm-view-item-set-summary (item val)
  (setf (nth 1 item) val))
(defun howm-view-item-set-offset (item val)
  (setf (nth 3 item) val))
(defun howm-view-item-set-home (item val)
  (setf (nth 4 item) val))
(defun howm-view-item-set-privilege (item val)
  (setf (nth 5 item) val))

(defun howm-view-item-basename (item &optional nonempty)
  (let* ((f (howm-view-item-filename item))
         (b (file-name-nondirectory f)))
    (if (and (string= b "") nonempty)
        f
      b)))
(defun howm-view-item-exclude-p (name-checker item)
  (and (not (howm-view-item-privilege item))
       (funcall name-checker (howm-view-item-filename item))))

;;; summary

(defvar *howm-view-summary-nop* nil)

(defun howm-view-summary (&optional name item-list)
  (when (null name)
    (setq name howm-view-name)
    (setq item-list howm-view-item-list))
  (if (null item-list)
      (progn (or *howm-view-summary-nop*
                 (message "No match"))
             nil)
    (let ((d default-directory))
      (howm-view-setup-buffer howm-view-summary-name name item-list)
      (setq default-directory d)
      (if *howm-view-summary-nop*
          t  ;; do nothing
        (howm-view-summary-subr name item-list)))))

(defun howm-view-summary-subr (name item-list)
    (progn
      (howm-view-summary-mode)
      (howm-view-summary-show item-list)
      (howm-view-summary-check t)
      (font-lock-mode 1)
      (when *howm-view-font-lock-keywrods*
        (setq font-lock-keywords *howm-view-font-lock-keywrods*))
      (setq font-lock-keywords-only t)
      (setq font-lock-keywords-case-fold-search
            howm-view-grep-ignore-case-option)
      (font-lock-fontify-block)
      t))

(defun howm-view-summary-show (item-list)
  (setq buffer-read-only nil)
  (erase-buffer)
  (mapc 'howm-view-summary-show-item item-list)
  (set-buffer-modified-p nil)
  (setq buffer-read-only t
        truncate-lines t)
  (goto-char (point-min))
  (setq howm-view-summary-last-line -777))

(defun howm-view-summary-show-item (item)
  (let* ((f (howm-view-item-basename item t))
         (h (format howm-view-summary-format f)))
    (insert h (howm-view-item-summary item) "\n")))

(defun howm-view-summary-mode ()
  "memo viewer (summary mode)
key	binding
---	-------
\\[howm-view-summary-open]	Open file
\\[next-line]	Next item
\\[previous-line]	Previous item
\\[howm-view-pop-or-scroll-other-window]	Pop and scroll contents
\\[scroll-other-window-down]	Scroll contents
\\[howm-view-scroll-other-window]	Scroll contents one line
\\[howm-view-scroll-other-window-down]	Scroll contents one line
\\[howm-view-summary-to-contents]	Concatenate all contents
\\[howm-view-filter-uniq]	Remove duplication of same file
\\[howm-view-summary-shell-command]	Execute command in inferior shell

\\[delete-other-windows]	Delete contents window
\\[howm-view-pop-window]	Pop contents window
\\[howm-view-toggle-window]	Toggle contents window

\\[howm-view-filter]	Filter (by date, contents, etc.)
\\[howm-view-filter-by-contents]	Search (= filter by contents)
\\[howm-view-sort]	Sort (by date, summary line, etc.)
\\[howm-view-sort-reverse]	Reverse order
\\[howm-view-dired]	Invoke Dired-X
\\[describe-mode]	This help
\\[howm-view-kill-buffer]	Quit
"
;; \\[howm-view-invoke-action-lock]	Invoke action-lock
;   (kill-all-local-variables)
  (setq major-mode 'howm-view-summary-mode
        mode-name "HowmS")
  (use-local-map howm-view-summary-mode-map)
  (make-local-variable 'font-lock-keywords)
  (make-local-variable 'font-lock-keywords-only)
  (make-local-variable 'font-lock-keywords-case-fold-search)
  (make-local-hook 'post-command-hook)
  (add-hook 'post-command-hook 'howm-view-post-command t t)
  (run-hooks 'howm-view-summary-mode-hook))

(defun howm-view-post-command ()
  (if *howm-view-summary-check*
      (howm-view-summary-check)
    (setq *howm-view-summary-check* t)))

(defun howm-view-summary-open (&optional reverse-delete-p)
  (interactive "P")
  (howm-view-summary-check t)
  (let* ((p (howm-view-persistent-p howm-view-summary-persistent))
         (persistent (if reverse-delete-p
                         (not p)
                       p)))
    (howm-view-summary-open-sub (not persistent))))

(defun howm-view-summary-open-sub (&optional kill)
  (interactive "P")
  (let ((b (howm-view-contents-buffer))
        (looking-at-str (buffer-substring-no-properties (point)
                                                        (line-end-position))))
;;         (howm-view-contents-persistent
;;          (howm-view-persistent-p howm-view-summary-persistent)))  ;; dirty!!
    (pop-to-buffer b)
    (howm-view-contents-open-sub kill)
    (end-of-line)
    (or (search-backward looking-at-str (line-beginning-position) t)
        (beginning-of-line))
    (run-hooks 'howm-view-open-hook)))

(defun howm-view-summary-current-item ()
  (let ((n (howm-view-line-number)))
    (nth (1- n) (howm-view-item-list))))

(defun howm-view-summary-check (&optional force)
  (let ((keep-one howm-view-keep-one-window))
    (when force
      (howm-view-refresh-window-configuration)
      (setq keep-one nil))
    (let ((n (howm-view-line-number))
          (howm-view-keep-one-window keep-one))
      (when (or (not (= n howm-view-summary-last-line))
                force)
        (setq howm-view-summary-last-line n)
        (let ((item (howm-view-summary-current-item)))
          (when (and item *howm-view-summary-check*)
            (howm-view-summary-update item force)))))))

(defvar howm-view-keep-one-window nil)
(defun howm-view-summary-update (item &optional new)
  (let ((execute t))
    (when (one-window-p)
      (cond (howm-view-keep-one-window (setq execute nil))
            ((eq howm-view-split-horizontally t) (split-window-horizontally))))
    (when execute
      (howm-view-summary-update-subr item new))))
(defun howm-view-summary-update-subr (item &optional new)
  (let* ((vbuf (howm-view-contents-buffer new))
	 (cwin (selected-window))
	 (pop-up-windows (or pop-up-windows howm-view-pop-up-windows))
         (fname (howm-view-item-filename item))
         (name (howm-view-name)))
    (pop-to-buffer vbuf)
    (howm-view-contents name (list item))
    (goto-char (point-min))
    (let ((home (howm-view-item-home item)))
      (when (numberp home)
        (goto-char home)
        (recenter howm-view-search-recenter))
      (message "View: %s" fname)
      (select-window cwin))))

(defun howm-view-pop-window ()
  (interactive)
  (let ((r (one-window-p)))
    (when r
      (howm-view-summary-check t))
    r))

(defun howm-view-pop-or-scroll-other-window ()
  (interactive)
  (or (howm-view-pop-window)
      (scroll-other-window)))

(defun howm-view-toggle-window ()
  (interactive)
  (or (howm-view-pop-window)
      (delete-other-windows)))

(defun howm-view-summary-to-contents ()
  (interactive)
  (let ((b (current-buffer)))
    (delete-other-windows)
    (set-buffer b)
    (let ((n (howm-view-line-number)))
      (howm-view-contents (howm-view-name) (howm-view-item-list))
      (goto-char (howm-view-contents-beginning (1- n))))))

;;; contents    

(defun howm-view-contents (name item-list)
  (if (null item-list)
      (message "Empty.")
    (let ((d (with-current-buffer (howm-view-summary-buffer)
               default-directory)))
      (howm-view-setup-buffer howm-view-contents-name name item-list)
      (setq default-directory d)
      (when (not (eq major-mode 'howm-view-contents-mode))
        (howm-view-contents-mode))
      (howm-view-contents-show item-list)
      (font-lock-fontify-block))))

(defvar howm-view-previous-file-name nil "For internal use") 
(defvar howm-view-previous-file-beg nil "For internal use") 
(defvar howm-view-previous-file-end nil "For internal use") 
(defun howm-view-contents-show (item-list)
  (setq buffer-read-only nil)
  (erase-buffer)
  (let ((howm-view-previous-file-name "") ;; dirty!
        (howm-view-previous-file-beg nil)
        (howm-view-previous-file-end nil))
    (setq howm-view-contents-end
          (mapcar 'howm-view-contents-show-item item-list)))
  (set-buffer-modified-p nil)
  (setq buffer-read-only t)
  (goto-char (point-min))
  )

(defun howm-view-insert-file/buffer-contents (file)
  (let ((b (get-file-buffer file)))
    (if (and b
             howm-view-watch-modified-buffer
             (not howm-view-use-grep))
        (insert-buffer b)
      (insert-file-contents file))))

(defun howm-view-contents-show-item (item)
  (let* ((fname (howm-view-item-filename item))
         (place (howm-view-item-place item))
         (header (format howm-view-header-format
                         (howm-abbreviate-file-name fname)))
         (viewer (howm-view-external-viewer fname))
         (done-p (if place
                     (and (string= fname howm-view-previous-file-name) ;; dirty!
                          (<= howm-view-previous-file-beg place)
                          (< place howm-view-previous-file-end))
                   (string= fname howm-view-previous-file-name))))
    (when (not done-p)
      (let (b e c h)
        (insert header)
        (setq c (point))
        (insert (with-temp-buffer
                  (if viewer
                      (insert "%%% "
                              (cond ((stringp viewer) (format viewer fname))
                                    ((functionp viewer) (format "%s" viewer))
                                    (t "Unknown viewer"))
                              " %%%")
                    (howm-view-insert-file/buffer-contents fname))
                  (if place
                      (progn
                        (howm-view-set-place place)
                        (setq h (point))
                        (let ((r (howm-view-paragraph-region fname)))
                          (setq b (car r)
                                e (second r))))
                    (setq b (point-min)
                          e (point-max)
                          h b))
                  (howm-view-item-set-offset item (- c b))
                  (howm-view-item-set-home item (+ c (- b) h))
                  (setq howm-view-previous-file-name fname ;; dirty!
                        howm-view-previous-file-beg (howm-view-get-place b)
                        howm-view-previous-file-end (howm-view-get-place e))
                  (buffer-substring-no-properties b e)))))
    (point)))

(defun howm-view-paragraph-region (&optional filename)
  (when filename
    (howm-set-configuration-for-file-name filename))
  (let ((b (save-excursion
             (end-of-line)
             (re-search-backward howm-view-title-regexp
                                 nil 'to-limit)
             (line-beginning-position)))
        (e (save-excursion
             (end-of-line)
             (if (re-search-forward howm-view-title-regexp
                                    nil 'to-limit)
                 (forward-line -1)
               (goto-char (point-max)))
             (end-of-line)
             (while (and (looking-at "^$")
                         (= (forward-line -1) 0)) ;; successful
               nil) ;; dummy
             (end-of-line)
             (point))))
    (list b e)))

(defun howm-view-contents-mode ()
  "memo viewer (contents mode)
key	binding
---	-------
\\[howm-view-contents-open]	Open file
\\[next-line]	Next line
\\[previous-line]	Previous line
\\[scroll-up]	Scroll up
\\[scroll-down]	Scroll down
\\[howm-view-scroll-up]	Scroll one line up
\\[howm-view-scroll-down]	Scroll one line down
\\[howm-view-contents-to-summary]	Summary
\\[howm-view-contents-goto-next-item]	Next item
\\[howm-view-contents-goto-previous-item]	Previous item

\\[howm-view-filter]	Filter (by date, contents, etc.)
\\[howm-view-filter-by-contents]	Search (= filter by contents)
\\[howm-view-sort]	Sort
\\[howm-view-sort-reverse]	Reverse order
\\[howm-view-dired]	Invoke Dired-X
\\[describe-mode]	This help
\\[howm-view-kill-buffer]	Quit
"
  "Not yet."
;   (kill-all-local-variables)
  (setq major-mode 'howm-view-contents-mode
        mode-name "HowmC")
  (use-local-map howm-view-contents-mode-map)
  (make-local-variable 'font-lock-keywords)
  (make-local-variable 'font-lock-keywords-only)
  (make-local-variable 'font-lock-keywords-case-fold-search)
  (cheat-font-lock-mode howm-view-font-lock-silent)
  (let ((ck `((,howm-view-header-regexp (0 howm-view-hilit-face))))
        ;; extremely dirty!! [2003/10/06 21:08]
        (sk (or (with-current-buffer (howm-view-summary-buffer)
                  font-lock-keywords)
                *howm-view-font-lock-keywrods*)))
    (cheat-font-lock-merge-keywords howm-view-contents-font-lock-keywords
                                    sk ck)
    (setq font-lock-keywords-only t)
    (setq font-lock-keywords-case-fold-search
          howm-view-grep-ignore-case-option)
;;    (font-lock-fontify-block)
    (run-hooks 'howm-view-contents-mode-hook)))

(defun howm-view-contents-open (&optional reverse-delete-p)
  (interactive "P")
  (let* ((p (howm-view-persistent-p howm-view-contents-persistent))
         (persistent (if reverse-delete-p
                         (not p)
                       p)))
    (howm-view-contents-open-sub (not persistent))))

(defun howm-view-contents-open-sub (&optional kill)
  (let* ((n (howm-view-contents-item-number (point)))
         (item (nth n (howm-view-item-list)))
         (file (howm-view-item-filename item))
         (offset (howm-view-item-offset item))
         (pos (- (point) offset))
         (viewer (howm-view-external-viewer file)))
    (when kill
      (howm-view-kill-buffer))
    (when (howm-view-item-privilege item)
      (howm-view-restore-window-configuration)) ;; force without mode check
    (run-hooks 'howm-view-before-open-hook)
    (if viewer
        (howm-view-call-external-viewer viewer file)
      (progn
        (find-file file)
        (howm-view-set-mark-command)
        (goto-char pos)
        (recenter howm-view-open-recenter)))
    (run-hooks 'howm-view-open-hook)))

(defun howm-view-contents-item-number (position)
  (let ((rest howm-view-contents-end)
        (n 0))
    (while (and rest (<= (car rest) position))
      (setq rest (cdr rest)
            n (+ n 1)))
    (min n (1- (length howm-view-contents-end))))) ;; for the last line

(defun howm-view-contents-beginning (n)
  (nth n (cons 1 howm-view-contents-end)))

(defun howm-view-contents-to-summary ()
  (interactive)
  (let ((n (howm-view-contents-item-number (point))))
    (howm-view-summary (howm-view-name) (howm-view-item-list))
    (goto-line (1+ n)))) ;; top = 1 for goto-line

(defun howm-view-contents-goto-next-item (&optional n)
  (interactive "p")
  (let* ((c (point))
         ;; remember that howm-view-contents-end has duplicats
         (stops (remove-duplicates (sort `(1
                                           ,c
                                           ,@(copy-sequence howm-view-contents-end))
                                         #'<)))
         (pos (position c stops))
         (new (+ pos n)))
    (cond ((< new 0)
           (goto-char (point-min))
           (error "Beginning of buffer"))
          ((>= new (length stops))
           (goto-char (point-max))
           (error "End of buffer"))
          (t
           (goto-char (nth new stops))))))

(defun howm-view-contents-goto-previous-item (&optional n)
  (interactive "p")
  (howm-view-contents-goto-next-item (- n)))

;;; misc.

(defun howm-view-setup-buffer (name-format name item-list)
  (switch-to-buffer (howm-view-get-buffer name-format name))
  (setq howm-view-name name
        howm-view-item-list item-list))
(defun howm-view-contents-buffer (&optional new)
  (howm-view-get-buffer howm-view-contents-name nil new))
(defun howm-view-summary-buffer (&optional new)
  (howm-view-get-buffer howm-view-summary-name nil new))
(defun howm-view-get-buffer (name-format &optional name new)
  (let* ((bufname (format name-format (or name (howm-view-name))))
         (buf (get-buffer bufname)))
    (when (and new buf)
      (kill-buffer buf))
    (get-buffer-create bufname)))

(defun howm-view-line-number (&optional pos)
  (save-excursion
    (when pos
      (goto-char pos))
    (let ((raw (count-lines (point-min) (point))))
      (if (bolp)
          (+ raw 1)
        raw))))

(defun howm-view-set-mark-command ()
  (set-mark-command nil)
  (howm-deactivate-mark))

(defun howm-view-kill-buffer ()
  (interactive)
  (let ((s (howm-view-summary-buffer))
        (c (howm-view-contents-buffer)))
    (kill-buffer s)
    (kill-buffer c)
    (howm-view-restore-window-configuration)))

(defun howm-view-name () howm-view-name)
(defun howm-view-item-list () howm-view-item-list)

(defun howm-view-file-list ()
  (remove-duplicates (mapcar #'howm-view-item-filename (howm-view-item-list))
                     :test #'string=))

(defun howm-view-xtime (file x)
  (let* ((a (file-attributes file))
         (n (cdr (assoc x '((a . 4) (m . 5) (c . 6)))))
         (ti (nth n a)))
    (howm-view-time-to-string ti)))

(defun howm-view-time-to-string (ti)
  (format-time-string "%Y%m%d-%H%M%S" ti))

(defun howm-view-string> (a b)
  (string< b a))

(defun howm-view-string<= (a b)
  (not (string< b a)))

;; 'Place' is line number at now
(defun howm-view-set-place (place)
  (goto-line place))
(defun howm-view-get-place (&optional point)
  (howm-view-line-number point))

(defun howm-view-scroll-up (count)
  (interactive "p")
  (scroll-up count))
(defun howm-view-scroll-down (count)
  (interactive "p")
  (scroll-down count))
(defun howm-view-scroll-other-window (count)
  (interactive "p")
  (scroll-other-window count))
(defun howm-view-scroll-other-window-down (count)
  (interactive "p")
  (scroll-other-window-down count))

(defun howm-view-persistent-p (z)
  (if (functionp z)
      (funcall z)
    z))

;;; dir

(defun howm-view-directory (dir &optional recursive-p)
  (let* ((files (if recursive-p
                    (howm-view-files dir)
                  (directory-files dir t)))
         (summary "")
         (items (mapcar (lambda (f)
                          (howm-view-make-item f summary))
                        files)))
    (howm-view-summary "" items)
    ;; sloppy!
    ;; (for old ruby-mode.el which sets global value of font-lock-keywords)
    (setq font-lock-keywords nil)
    ))

;; list files recursively
(defun howm-view-files (dir)
  (cond ((file-directory-p dir) (howm-view-files-sub dir))
        ((file-exists-p dir) (list dir))
        (t nil)))

(defun howm-view-files-sub (dir)
  (setq dir (expand-file-name dir))
  (mapcan (lambda (f)
            (cond
             ((file-directory-p f) (if (howm-subdirectory-p f dir t)
                                       (howm-view-files f)
                                     nil)) ;; exclude "." & ".."
             ((file-regular-p f) (list f))
             (t nil)))
          (directory-files dir t)))

;; (defun howm-view-files-sub (dir)
;;   (setq dir (expand-file-name dir))
;;   (mapcan (lambda (f)
;;             (cond ((file-directory-p f) (let ((d (expand-file-name f)))
;;                                           ;; exclude "." & ".."
;;                                           (if (string< dir d)
;;                                               (howm-view-files d)
;;                                             nil)))
;;                   ((file-regular-p f) (list f))
;;                   (t nil)))
;;           (directory-files dir t)))

;;; filter

(defun howm-view-filter ()
  (interactive)
  (let* ((table howm-view-filter-methods)
         (command (completing-read "filter by: " table nil t)))
    (call-interactively (cdr (assoc command table)))))

(defvar howm-view-filter-uniq-prev "")
(defun howm-view-filter-uniq ()
  (interactive)
  (howm-view-filter-general
   (lambda (item)
     (let ((f (howm-view-item-filename item)))
       (prog1
           (not (string= f howm-view-filter-uniq-prev))
         (setq howm-view-filter-uniq-prev f))))))

(defun howm-view-filter-by-name (&optional remove-p regexp)
  (interactive "P")
  (howm-view-filter-by-name/summary 'howm-view-item-basename
                                    remove-p regexp))

(defun howm-view-filter-by-summary (&optional remove-p regexp)
  (interactive "P")
  (howm-view-filter-by-name/summary 'howm-view-item-summary
                                    remove-p regexp))

(defun howm-view-filter-by-name/summary (accessor remove-p regexp)
  (let* ((r (or regexp (read-from-minibuffer "Regexp: "))))
    (howm-view-filter-general (lambda (item)
                                (string-match r (funcall accessor item)))
                              remove-p)))

(defun howm-view-filter-by-date (&optional remove-p)
  (interactive "P")
  (let* ((r (howm-view-ask-time-range))
         (from (car r))
         (to (second r))
         (form (howm-view-file-name-format))
         (fts (mapcar (lambda (x)
                        (file-name-nondirectory (format-time-string form x)))
                      (list from to)))
         (fs (car fts))
         (ts (second fts)))
    (howm-view-filter-general
     (lambda (item)
       (let ((cs (howm-view-item-basename item)))
         (and (howm-view-string<= fs cs) (howm-view-string<= cs ts))))
     remove-p)
    ))

(defun howm-view-filter-by-reminder (&optional remove-p)
  (interactive "P")
  (let* ((r (howm-view-ask-time-range))
         (from (car r))
         (to (second r))
         (from-str (format-time-string howm-date-format from))
         (to-str (format-time-string howm-date-format to))
         (reg (howm-reminder-regexp howm-reminder-types)))
    (howm-view-filter-general
     (lambda (item)
       (let ((s (howm-view-item-summary item)))
         (and (string-match reg s)
              (let* ((x (match-string-no-properties 0 s)) ;; [2004-02-07]@
                     (d (and (string-match howm-date-regexp x)
                             (match-string-no-properties 0 x)))) ;; [2004-02-07]
                (and (howm-view-string<= from-str d)
                     (howm-view-string<= d to-str))))))
     remove-p)
    ))

(defun howm-view-file-name-format ()
  howm-file-name-format) ;; defined in howm-common.el

;; needless
; (defun howm-view-filter-by-ctime (&optional remove-p)
;   (interactive "P")
;   (howm-view-filter-by-xtime 'c remove-p))

(defun howm-view-filter-by-mtime (&optional remove-p range)
  (interactive "P")
  (howm-view-filter-by-xtime 'm remove-p range))

(defun howm-view-filter-by-xtime (x remove-p &optional range)
  (let* ((r (or range (howm-view-ask-time-range)))
         (from (car r))
         (to (second r))
         (fs (howm-view-time-to-string from))
         (ts (howm-view-time-to-string to)))
    (howm-view-filter-general
     (lambda (item)
       (let* ((cs (howm-view-xtime (howm-view-item-filename item) x)))
         (and (howm-view-string<= fs cs) (howm-view-string<= cs ts))))
     remove-p)))

(defun howm-view-ask-time-range ()
  (let* ((now (current-time))
         (from (howm-view-ask-time "From" now t))
         (to (howm-view-ask-time "To" from)))
    (list from to)))

(defvar howm-view-min-year 1950)
(defvar howm-view-max-year 2030)
(defun howm-view-ask-time (prompt default &optional from-p)
  (let* ((z (decode-time default))
         (yd (nth 5 z))
         (md (nth 4 z))
         (dd (nth 3 z)))
    (let (y0 m0 d0 hour0 min0 sec0)
      (if from-p
          (setq y0 howm-view-min-year m0 1 d0 1
                hour0 0 min0 0 sec0 0)
        (setq y0 howm-view-max-year m0 12 d0 'last-day-of-month
              hour0 24 min0 0 sec0 0))
      (let ((y (howm-ask-time-sub prompt "year" yd)))
        (if (null y)
            (howm-view-encode-time sec0 min0 hour0 d0 m0 y0)
          (let ((m (howm-ask-time-sub prompt "month" md)))
            (if (null m)
                (howm-view-encode-time sec0 min0 hour0 d0 m0 y)
              (let ((d (or (howm-ask-time-sub prompt "date" dd) d0)))
                (howm-view-encode-time sec0 min0 hour0 d m y)))))))))

(defun howm-ask-time-sub (prompt ymd default)
  (let* ((raw (read-from-minibuffer (format "%s %s (* = no limit) [%d]: "
                                            prompt ymd  default)))
         (n (if (string= raw "")
                default
              (string-to-number raw))))
    (if (= n 0)
        nil
      n)))

(defun howm-view-encode-time (sec min hour d m y)
  (when (eq d 'last-day-of-month)
    (setq m (+ m 1))
    (setq d -1))
  (encode-time sec min hour d m y))

(defun howm-view-filter-general (pred &optional remove-p)
  (let* ((item-list (howm-view-item-list))
         (filtered (if remove-p
                      (remove-if pred item-list)
                      (remove-if-not pred item-list))))
    (howm-view-summary-rebuild filtered)))

(defun howm-view-summary-rebuild (item-list)
  (let ((name (howm-view-name)))
    (howm-view-summary name item-list)))

(defun howm-view-filter-by-contents (str)
  (interactive "sSearch in result (grep): ")
  (let* ((orig (howm-view-name))
         (name (if (string= orig "")
                   str
                 (format "%s&%s" orig str)))
         (file-list (howm-view-file-list)))
    (howm-view-search str file-list name)))

(defun howm-view-search (str file-list &optional
                             name summarizer fixed-p hilit-keywords)
  (setq name (or name str))
  (when howm-view-update-search-ring
    (isearch-update-ring str (not fixed-p)))
  (let* ((items (howm-view-search-items str file-list summarizer fixed-p))
         (kw (or hilit-keywords
                 `((,(regexp-quote str) . howm-view-hilit-face))))
         )
    (let* ((f (expand-file-name str)))
      (when (file-exists-p f)
        (let ((fi (howm-view-make-item f)))
          (howm-view-item-set-privilege fi t)
          (setq items (cons fi items)))))
    (let ((*howm-view-font-lock-keywrods* kw)) ;; dirty!
      (howm-view-summary name items))))

(defun howm-view-search-items (str file-list &optional summarizer fixed-p)
  (let ((found (howm-view-grep str file-list fixed-p))
        (summarizer (or summarizer
                        (lambda (file place content)
                          (string-match "^ *\\(.*\\)" content)
                          (match-string-no-properties 1 content)))))
    (mapcar (lambda (z)
              (let ((file (car z))
                    (place (second z))
                    (content (third z)))
                (howm-view-make-item file
                                     (funcall summarizer
                                              file place content)
                                     place)))
            found)))

(defvar *howm-view-force-case-fold-search* nil) ;; dirty!
(defvar howm-view-grep-log-file nil)
(defun howm-view-grep (str file-list &optional fixed-p)
  (when howm-view-grep-log-file
    (save-excursion
      (find-file howm-view-grep-log-file)
      (goto-char (point-max))
      (insert (format "%s %s\n"
                      (format-time-string "%Y-%m-%d-%H%M%S" (current-time))
                      str))
      (basic-save-buffer)
      (bury-buffer)))
  (when (stringp file-list)
    (setq file-list (list file-list)))
  (when *howm-view-force-case-fold-search*
    (setq str (downcase str)))
  (if howm-view-use-grep
      (howm-view-real-grep str file-list fixed-p)
    (howm-view-fake-grep str file-list fixed-p)))

(defun howm-view-real-grep (str file-list &optional fixed-p)
  "Call grep and parse its result.
'((file line-number line) (file line-number line) ...)
"
  (let ((grep-command (if fixed-p
                          howm-view-fgrep-command
                        howm-view-grep-command))
        (opt (list howm-view-grep-option))
        (eopt (and howm-view-grep-expr-option
                   (list howm-view-grep-expr-option))))
    (when (and (not (let ((case-fold-search nil))
                      (string-match "[A-Z]" str)))
               howm-view-grep-ignore-case-option)
      (setq opt (cons howm-view-grep-ignore-case-option opt)))
    (with-temp-buffer
      (let* ((fs (mapcar #'expand-file-name file-list))
             (lines (howm-view-call-process grep-command
                                            `(,@opt ,@eopt ,str ,@fs)))
             (parsed (mapcar 'howm-view-grep-parse-line lines)))
        (remove nil parsed)))))

(defun howm-view-fake-grep (str file-list &optional fixed-p)
  (let ((search-backward-f (if fixed-p
                              #'search-backward
                            #'re-search-backward)))
    (mapcan (lambda (file)
              (howm-view-fake-grep-file str file search-backward-f))
            (mapcan #'howm-view-files file-list))))

(defun howm-view-fake-grep-file (str file search-backward-f)
  (let ((b (get-file-buffer file)))
    (if (and b howm-view-watch-modified-buffer)
        (with-current-buffer b
          (howm-view-fake-grep-current-buffer str file search-backward-f))
      (with-temp-buffer
        (insert-file-contents file)
        (howm-view-fake-grep-current-buffer str file search-backward-f)))))

(defun howm-view-fake-grep-current-buffer (str file search-backward-f)
  (save-excursion
    (goto-char (point-max))
    (let ((found nil))
      (while (funcall search-backward-f str nil t)
        (beginning-of-line)
        (setq found
              (cons (list file
                          (howm-view-get-place)
                          (buffer-substring-no-properties (point)
                                                          (line-end-position)))
                    found)))
      found)))

(defvar howm-view-call-process-last-command nil)  ;; for debug
(defun howm-view-call-process (command args &optional expected-return-value)
  (setq howm-view-call-process-last-command (cons command args))
  (with-temp-buffer
    (let ((r (howm-view-call-process-here command args)))
      (when (and expected-return-value
                 (not (= expected-return-value r)))
        (error "Process returns %s instead of expected %s."
               r expected-return-value))
      (howm-view-buffer-lines))))

(defun howm-view-call-process-here (command args)
  (let ((a `(,command nil (t nil) nil ,@args)))
    (apply #'call-process a)))

(defun howm-view-buffer-lines (&optional buf)
  (save-excursion
    (when buf
      (set-buffer buf))
    (split-string (buffer-substring (point-min) (point-max)) "\n")))

(defun howm-view-grep-parse-line (line)
  (if (string-match "^\\(\\([a-zA-Z]:/\\)?[^:]*\\):\\([0-9]*\\):\\(.*\\)$"
                    line)
      (let ((file (match-string 1 line))
            (line (string-to-number (match-string 3 line)))
            (content (match-string 4 line)))
        (list file line content))
    nil))

;;; sort

(defun howm-view-sort ()
  (interactive)
  (let* ((table howm-view-sort-methods)
         (command (completing-read "sort by: " table nil t)))
    (call-interactively (cdr (assoc command table)))))

(defun howm-view-sort-by-name (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-basename #'string< reverse-p))

(defun howm-view-sort-by-name-match (&optional reverse-p regexp path-p)
  (interactive "P")
  (howm-view-sort-by-general-match (if path-p
                                       #'howm-view-item-filename
                                     #'howm-view-item-basename)
                                   reverse-p regexp))

(defvar howm-view-sort-by-date-ignore-regexp "^[a-zA-Z]")
(defun howm-view-sort-by-date (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-basename #'string<
                          reverse-p)
  (howm-view-sort-by-general-match #'howm-view-item-basename t
                                   howm-view-sort-by-date-ignore-regexp))
(defun howm-view-sort-by-reverse-date (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-by-date (not reverse-p)))

(defun howm-view-sort-by-summary (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-summary #'string< reverse-p))

(defun howm-view-sort-by-summary-match (&optional reverse-p regexp)
  (interactive "P")
  (howm-view-sort-by-general-match #'howm-view-item-summary
                                   reverse-p regexp))

(defun howm-view-sort-by-general-match (picker &optional reverse-p regexp)
  (let ((r (or regexp (read-from-minibuffer "Regexp: ")))
        (howm-view-s-b-g-m-matched nil)) ;; need unique name?? :-(
    (howm-view-sort-general (lambda (item)
                              (if (string-match r
                                                (funcall picker item))
                                  (progn
                                    (setq howm-view-s-b-g-m-matched t)
                                    1)
                                0))
                            #'>
                            reverse-p)
    howm-view-s-b-g-m-matched))

(defun howm-view-sort-by-reminder (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general (lambda (item)
                            (let ((s (howm-view-item-summary item))
                                  (r (howm-reminder-regexp howm-reminder-types))
                                  (max-str (format-time-string
                                            howm-reminder-today-format
                                            (encode-time 59 59 23 31 12
                                                         howm-view-max-year))))
                              (if (string-match r s)
                                  (match-string-no-properties 0 s)
                                max-str)))
                          #'string< reverse-p))

;; nonsense
; (defun howm-view-sort-by-atime (&optional reverse-p)
;   (interactive "P")
;   (howm-view-sort-by-xtime 'a reverse-p))

;; needless
; (defun howm-view-sort-by-ctime (&optional reverse-p)
;   (interactive "P")
;   (howm-view-sort-by-xtime 'c reverse-p))

(defun howm-view-sort-by-mtime (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-by-xtime 'm reverse-p))

(defun howm-view-sort-by-xtime (x reverse-p)
  (howm-view-sort-general (lambda (item)
                            (howm-view-xtime (howm-view-item-filename item)
                                             x))
;                           #'>
;                           reverse-p
                          #'howm-view-string>
                          reverse-p
                          ))

(defun howm-view-sort-general (evaluator comparer &optional reverse-p)
  (let* ((howm-view-s-g-comparer comparer) ;; need unique name?? :-(
         (cmp (if reverse-p
                  (lambda (a b) (funcall howm-view-s-g-comparer b a))
                howm-view-s-g-comparer))
         (sorted (howm-view-sort-items evaluator cmp
                                       (howm-view-item-list))))
    (howm-view-summary (howm-view-name) sorted)))

(defun howm-view-sort-items (evaluator comparer item-list)
  (let* ((orig (mapcar (lambda (item) (cons (funcall evaluator item) item))
                       item-list))
         (sorted (sort orig (lambda (x y)
                              (funcall comparer (car x) (car y))))))
    (mapcar #'cdr sorted)))

(defun howm-view-sort-reverse ()
  (interactive)
  (howm-view-summary (howm-view-name)
                    (reverse (howm-view-item-list))))

;;; Dired-X

(defvar howm-view-dired-buffer-name "*howm-dired*")
(defvar howm-view-dired-ls-command "ls")
(defvar howm-view-dired-ls-options '("-l"))

(defun dired-virtual (dir)
  (howm-inhibit-warning-in-compilation))

(defun howm-view-dired ()
  (interactive)
  (require 'dired-x)
  (when (not (member major-mode
                     '(howm-view-summary-mode howm-view-contents-mode)))
    (error "Invalid mode for this command."))
;;   ;; bug in emacs-21.3.50?
;;   (when (not (fboundp 'dired-insert-headerline))
;;     (defun dired-insert-headerline (dir);; also used by dired-insert-subdir
;;       ;; Insert DIR's headerline with no trailing slash, exactly like ls
;;       ;; would, and put cursor where dired-build-subdir-alist puts subdir
;;       ;; boundaries.
;;       (save-excursion (insert "  " (directory-file-name dir) ":\n"))))
  (let* ((files (mapcar (lambda (item)
                          (file-relative-name (howm-view-item-filename item)))
                        howm-view-item-list))
         (args (append howm-view-dired-ls-options files))
         (a `((howm-view-summary-mode . ,howm-view-summary-persistent)
              (howm-view-contents-mode . ,howm-view-contents-persistent)))
         (p (howm-view-persistent-p (cdr (assoc major-mode a)))))
    (if p
        (howm-view-restore-window-configuration)
      (howm-view-kill-buffer))
    (switch-to-buffer (get-buffer-create howm-view-dired-buffer-name))
    (setq buffer-read-only nil)
    (erase-buffer)
    (howm-view-call-process-here howm-view-dired-ls-command args)
    (set-buffer-modified-p nil)
    (dired-virtual default-directory)))

;;; shell

(defvar howm-view-summary-shell-hist '("ls -l FILE" "FILE"))
(defvar howm-view-summary-shell-last-file "FILE")
(defun howm-view-summary-shell-command ()
  (interactive)
  (when (not (member major-mode
                     '(howm-view-summary-mode)))
    (error "Invalid mode for this command."))
  (let* ((n (howm-view-line-number))
         (item (nth (1- n) (howm-view-item-list)))
         (file (howm-abbreviate-file-name (howm-view-item-filename item)))
         (last-reg (regexp-quote howm-view-summary-shell-last-file)))
    (setq howm-view-summary-shell-hist
          (mapcar (lambda (h)
                    (replace-regexp-in-string last-reg file h t))
                  howm-view-summary-shell-hist))
    (setq howm-view-summary-shell-last-file file)
    (let* ((default (car howm-view-summary-shell-hist))
           (c (read-string "command: "
                           (cons default 0)
                           '(howm-view-summary-shell-hist . 1))))
      (shell-command c))
    (let ((item-list (remove-if (lambda (item)
                                  (not (file-exists-p
                                        (howm-view-item-filename item))))
                                howm-view-item-list)))
      (setq *howm-view-summary-check* nil) ;; dirty
      (howm-view-summary howm-view-name item-list)
      (goto-line n)
      (save-selected-window
        (let ((b (get-buffer "*Shell Command Output*")))
          (cond ((not (howm-buffer-empty-p b))
                 (switch-to-buffer-other-window b))
                ((eq item (howm-view-summary-current-item))
                 nil)
                (t (progn
                     (setq *howm-view-summary-check* t) ;; dirty
                     (howm-view-summary-check t))))))
      )))

;;; xemacs

(defun howm-xemacsp ()
  (and (boundp 'xemacsp) xemacsp))

(defun howm-deactivate-mark ()
  (if (howm-xemacsp)
      (zmacs-deactivate-region)
    (deactivate-mark)))

(defun howm-try-require (feature)
  (and (locate-library (symbol-name feature))
       (require feature)))

;;; external viewer

(defvar howm-view-external-viewer-assoc nil)
(defvar howm-view-use-mailcap
  (and (howm-try-require 'mailcap)
       (fboundp 'mailcap-parse-mailcaps)
       (fboundp 'mailcap-parse-mimetypes)))
(defvar howm-view-open-by-myself  ;; regexp list
  '("text/.*" "application/emacs-lisp"))
(when howm-view-use-mailcap
  (require 'mailcap)
  (mailcap-parse-mailcaps)
  (mailcap-parse-mimetypes))

(defun howm-view-external-viewer (file)
  (let ((dir-viewer (and (file-directory-p file) #'howm-view-directory-viewer))
        (viewer (cdr (assoc-if (lambda (reg) (string-match reg file))
                               howm-view-external-viewer-assoc))))
    (or dir-viewer viewer
        (and howm-view-use-mailcap
             (let* ((ext (if (string-match "\\.[^\\.]+$" file)
                             (match-string 0 file)
                           ""))
                    (type (mailcap-extension-to-mime ext))
                    (type-match (lambda (r) (string-match r type))))
               (cond ((null type) nil)
                     ((member-if type-match howm-view-open-by-myself) nil)
                     (t (mailcap-mime-info type))))))))

(defun howm-view-call-external-viewer (viewer file)
  (cond ((functionp viewer)
         (progn
           (find-file file)
           (funcall viewer)))
        ((stringp viewer)
         (start-process "howm-view-external-viewer" nil
                        shell-file-name
                        shell-command-switch
                        (format viewer file)))
        (t (error "Unknown viewer %s for type %s." viewer file))))

(defun howm-view-directory-viewer ()
  t) ;; NOP

;;; provide

(provide 'howm-view)

;;; howm-view.el ends here
